import birl.{type Time}
import gleam/erlang/process
import gleam/io
import gleam/list
import gleam/option.{type Option, None, Some}
import shore
import shore/key
import shore/layout
import shore/style
import shore/ui
import spacetraders_api.{type ServerStatus}
import spacetraders_client/env.{type Env}
import spacetraders_client/pretty_print
import spacetraders_sdk.{type AgentToken, type ApiError, type ApiResponse}

pub fn print_general_info(agent_token: AgentToken) {
  let assert Ok(account) = spacetraders_api.get_account(agent_token)
  pretty_print.account(account)
  io.println("")

  let assert Ok(agent) = spacetraders_api.get_agent(agent_token)
  pretty_print.agent(agent)
  io.println("")

  let assert Ok(server_status) = spacetraders_api.get_server_status()
  pretty_print.server_status(server_status, True, True)
  io.println("")
}

pub fn print_ships(agent_token: AgentToken) {
  let assert Ok(ships) =
    spacetraders_api.list_ships(agent_token, option.None, option.None)
  list.each(ships.data, fn(ship) {
    pretty_print.ship(ship)
    io.println("")
  })
}

pub fn print_contracts(agent_token: AgentToken) {
  let assert Ok(contracts) =
    spacetraders_api.list_contracts(agent_token, option.None, option.None)
  list.each(contracts.data, fn(contract) {
    pretty_print.contract(contract)
    io.println("")
  })
}

pub fn main() -> Nil {
  let exit = process.new_subject()
  let assert Ok(_actor) =
    shore.spec(
      init:,
      update:,
      view:,
      exit:,
      keybinds: shore.default_keybinds(),
      redraw: shore.on_timer(16),
    )
    |> shore.start
  process.receive_forever(exit)
}

pub opaque type LastServerStatus {
  Loading
  Loaded(response: ApiResponse(ServerStatus))
  Refreshing(prev_response: ApiResponse(ServerStatus))
  Refreshed(
    prev_response: ApiResponse(ServerStatus),
    response: ApiResponse(ServerStatus),
  )
}

pub opaque type Model {
  Model(
    env: Env,
    errors: List(ApiError),
    last_request: Option(Time),
    last_server_status: LastServerStatus,
  )
}

pub opaque type Msg {
  NoOp
  DismissErrors
  GetServerStatus
  GotServerStatus(response: ApiResponse(ServerStatus))
}

fn init() -> #(Model, List(fn() -> Msg)) {
  let assert Ok(env) = env.load_dotenv()
  let model =
    Model(env:, errors: [], last_request: None, last_server_status: Loading)
  let cmds = [fn() { GetServerStatus }]
  #(model, cmds)
}

fn update(model: Model, msg: Msg) -> #(Model, List(fn() -> Msg)) {
  case msg {
    NoOp -> #(model, [])
    DismissErrors -> #(Model(..model, errors: []), [])
    GetServerStatus -> #(
      Model(..model, last_server_status: case model.last_server_status {
        Loading -> Loading
        Loaded(prev_response)
        | Refreshing(prev_response)
        | Refreshed(response: prev_response, ..) -> Refreshing(prev_response:)
      }),
      [fn() { GotServerStatus(spacetraders_api.get_server_status()) }],
    )
    GotServerStatus(response) -> {
      #(
        Model(
          ..model,
          last_request: Some(birl.now()),
          last_server_status: case model.last_server_status {
            Loading -> Loaded(response)
            Loaded(prev_response)
            | Refreshing(prev_response)
            | Refreshed(response: prev_response, ..) ->
              Refreshed(prev_response:, response:)
          },
        ),
        [],
      )
    }
  }
}

fn view(model: Model) -> shore.Node(Msg) {
  layout.grid(0, [style.Px(5), style.Fill], [style.Fill], [
    layout.cell(
      ui.box(
        [
          case model.last_server_status {
            Loaded(response: Ok(server_status))
            | Refreshed(response: Ok(server_status), ..) ->
              ui.text(server_status.status)
            Loading -> ui.text("Loading server status...")
            Refreshing(..) -> ui.text("Refreshing server status...")
            Loaded(response: Error(..)) ->
              ui.text_styled(
                "Failed to load server status",
                Some(style.Red),
                None,
              )
            Refreshed(response: Error(..), ..) ->
              ui.text_styled(
                "Failed to refresh server status",
                Some(style.Red),
                None,
              )
          },
          ui.br(),
          ui.row([ui.button("(R)efresh", key.Char("r"), GetServerStatus)]),
        ],
        Some("Server Status"),
      )
        |> ui.align(style.Center, _),
      #(0, 0),
      #(0, 0),
    ),
    layout.cell(
      ui.box(
        [ui.row([ui.button("(D)ismiss Errors", key.Char("d"), DismissErrors)])],
        Some("SpaceTraders"),
      ),
      #(1, 1),
      #(0, 0),
    ),
  ])
}
